package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.entity.BlockEntityType
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.items.CapabilityItemHandler
import net.minecraftforge.items.IItemHandler
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.block.RotatableMatteryBlock
import ru.dbotthepony.mc.otm.capability.*
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.plus
import ru.dbotthepony.mc.otm.core.toImpreciseFraction
import ru.dbotthepony.mc.otm.graph.storage.BasicStorageGraphNode
import ru.dbotthepony.mc.otm.graph.storage.StorageNetworkGraph
import ru.dbotthepony.mc.otm.orThrow
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.registry.MNames
import ru.dbotthepony.mc.otm.storage.IStorageTuple
import ru.dbotthepony.mc.otm.storage.ITEM_STORAGE
import ru.dbotthepony.mc.otm.storage.ItemStackWrapper
import ru.dbotthepony.mc.otm.unaryMinus

abstract class AbstractStorageImportExport<T>(
	blockType: BlockEntityType<*>,
	blockPos: BlockPos,
	blockState: BlockState,
	maxBatteryLevel: ImpreciseFraction = MAX_POWER,
	maxInput: ImpreciseFraction? = BlockEnergyStorageImpl.DEFAULT_MAX_IO,
	maxOutput: ImpreciseFraction? = maxInput
) : MatteryPoweredBlockEntity(blockType, blockPos, blockState) {
	final override val energy = WorkerEnergyStorage(this, maxBatteryLevel, maxInput, maxOutput)
	val cell = BasicStorageGraphNode(energy)

	private var valid = true

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		return if (valid && cap === MatteryCapability.STORAGE_NODE) {
			cell.get().cast()
		} else super.getCapability(cap, side)
	}

	override fun invalidateCaps() {
		super.invalidateCaps()
		cell.invalidate()
		valid = false
	}

	override fun reviveCaps() {
		super.reviveCaps()
		cell.revive()
		valid = true
	}

	override fun setRemoved() {
		super.setRemoved()
		cell.destroy(level)
		valid = false
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)

		if (p_155231_ is ServerLevel) {
			StorageNetworkGraph.discoverFull(this, cell.storageNode)
		}

		tickOnceServer(this::checkSurroundings)
	}

	protected var target: LazyOptional<T> = LazyOptional.empty()
	protected abstract val targetCapability: Capability<T>

	fun checkSurroundings() {
		target = getAndBind(
			target,
			level?.getBlockEntity(blockPos + blockState.getValue(RotatableMatteryBlock.FACING_FULL).normal),
			targetCapability,
			-blockState.getValue(RotatableMatteryBlock.FACING_FULL),
		) { tickOnceServer(this::checkSurroundings) }
	}

	companion object {
		val MAX_POWER = ImpreciseFraction(10_000)
	}
}

class StorageImporterBlockEntity(blockPos: BlockPos, blockState: BlockState) : AbstractStorageImportExport<IItemHandler>(MBlockEntities.STORAGE_IMPORTER, blockPos, blockState) {
	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu? {
		return null
	}

	private var lastSlot = 0
	private var nextTick = INTERVAL

	private val enoughEnergy get() = energy.batteryLevel >= ITEM_STORAGE.energyPerOperation

	override val targetCapability: Capability<IItemHandler>
		get() = CapabilityItemHandler.ITEM_HANDLER_CAPABILITY

	fun tick() {
		batteryChargeLoop()
		cell.tickEnergyDemanding()

		nextTick--

		if (nextTick <= 0 && target.isPresent && enoughEnergy) {
			val graph = cell.storageGraph ?: return
			val items = graph.getVirtualComponent(OverdriveThatMatters.INSTANCE.ITEM_STORAGE())

			val resolved = target.orThrow()

			if (lastSlot >= resolved.slots) {
				lastSlot = 0
			}

			val maxMove = energy.extractStepInner(ITEM_STORAGE.energyPerOperation, MAX_MOVE_PER_OPERATION, true)
			var extracted = resolved.extractItem(lastSlot, maxMove, true)

			if (extracted.isEmpty) {
				lastSlot++
			} else {
				val leftOver = items.insertStack(ItemStackWrapper(extracted), true)

				if (leftOver.count.toInt() != extracted.count) {
					extracted = resolved.extractItem(lastSlot, extracted.count - leftOver.count.toInt(), false)
					energy.extractStepInner(ITEM_STORAGE.energyPerOperation, extracted.count, false)
					items.insertStack(ItemStackWrapper(extracted), false)
				} else {
					nextTick += INTERVAL * 4
				}
			}
		}

		if (nextTick <= 0) {
			nextTick = INTERVAL
		}
	}

	companion object {
		const val MAX_MOVE_PER_OPERATION = 4
		private val MACHINE_NAME = TranslatableComponent("block.${OverdriveThatMatters.MOD_ID}.${MNames.STORAGE_IMPORTER}")
		private const val INTERVAL = 5
	}
}

class StorageExporterBlockEntity(blockPos: BlockPos, blockState: BlockState) : AbstractStorageImportExport<IItemHandler>(MBlockEntities.STORAGE_EXPORTER, blockPos, blockState) {
	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu? {
		return null
	}

	private var lastSlot = 0
	private var nextTick = INTERVAL
	private val enoughEnergy get() = energy.batteryLevel >= ITEM_STORAGE.energyPerOperation

	override val targetCapability: Capability<IItemHandler>
		get() = CapabilityItemHandler.ITEM_HANDLER_CAPABILITY

	private val exportStacks: List<IStorageTuple<ItemStackWrapper>>
		get() = cell.storageGraph?.getVirtualComponent(ITEM_STORAGE)?.getStacks() ?: emptyList()

	fun tick() {
		batteryChargeLoop()
		cell.tickEnergyDemanding()

		nextTick--

		if (nextTick <= 0 && target.isPresent && enoughEnergy) {
			val graph = cell.storageGraph ?: return
			val items = graph.getVirtualComponent(ITEM_STORAGE)

			val resolved = target.orThrow()

			if (lastSlot >= resolved.slots) {
				lastSlot = 0
			}

			var hit = false

			for (stack in exportStacks) {
				val exportAmountA = items.extractStack(stack.id, stack.stack.count.toInt().coerceAtMost(MAX_MOVE_PER_OPERATION).toImpreciseFraction(), true).count.toInt()

				if (exportAmountA == 0) {
					continue
				}

				var exportAmount = energy.extractStepInner(ITEM_STORAGE.energyPerOperation, exportAmountA, true)

				if (exportAmount == 0) {
					break
				}

				val leftover = resolved.insertItem(lastSlot, stack.stack.stack.also { it.count = exportAmount }, true)

				if (leftover.count != exportAmount) {
					hit = true
					exportAmount = items.extractStack(stack.id, ImpreciseFraction(exportAmount - leftover.count), false).count.toInt()
					resolved.insertItem(lastSlot, stack.stack.stack.also { it.count = exportAmount }, false)
					energy.extractStepInner(ITEM_STORAGE.energyPerOperation, exportAmount, false)
					break
				}
			}

			if (!hit) {
				lastSlot++
			}
		}

		if (nextTick <= 0) {
			nextTick = INTERVAL
		}
	}

	companion object {
		const val MAX_MOVE_PER_OPERATION = 4
		private val MACHINE_NAME = TranslatableComponent("block.${OverdriveThatMatters.MOD_ID}.${MNames.STORAGE_EXPORTER}")
		private const val INTERVAL = 5
	}
}
